from abc import ABC
from typing import Optional
import posixpath
from importlib.metadata import version
from decimal import Decimal

class BaseEngine(ABC):
    """
    Abstract base class for implementing different engine types.

    Attributes
    ----------
    SQLGLOT_DIALECT : str, optional
        Specifies the SQL dialect to be used by the engine when SQL transpiling
        is required. Default is None.
    REQUIRED_READ_ENDPOINT : str, optional
        Specifies `mount` or `abfss` if the engine only supports one endpoint. Default is None.
    REQUIRED_WRITE_ENDPOINT : str, optional
        Specifies `mount` or `abfss` if the engine only supports one endpoint. Default is None.

    Methods
    -------
    get_total_cores()
        Returns the total number of CPU cores available on the system.
    get_compute_size()
        Returns a formatted string with the compute size.
    append_array_to_delta(abfss_path: str, array: list)
        Appends a list of data to a Delta table at the specified path.
    """
    SQLGLOT_DIALECT = None
    REQUIRED_READ_ENDPOINT = None
    REQUIRED_WRITE_ENDPOINT = None
    SUPPORTS_SCHEMA_PREP = False
    _FABRIC_USD_COST_PER_VCORE_HOUR = 0.09  # cost in East US 2 as of July 2025
    
    def __init__(self):
        try:
            from IPython.core.getipython import get_ipython
            self.notebookutils = get_ipython().user_ns.get("notebookutils")
            self.is_fabric = True if self.notebookutils.runtime.context['productType'] == 'Fabric' else False
        except:
            self.is_fabric = False

        self.version: str = ''
        self.cost_per_vcore_hour: Optional[float] = None
                  
    def get_total_cores(self) -> int:
        """
        Returns the total number of CPU cores available on the system.
        """
        import os
        cores = os.cpu_count()
        return cores
    
    def get_compute_size(self) -> str:
        """
        Returns a formatted string with the compute size.
        """
        cores = self.get_total_cores()
        return f"{cores}vCore"
    
    def get_job_cost(self, duration_ms: int) -> Optional[Decimal]:
        """
        Returns the cost per hour for compute as a Decimal.
        
        If `cost_per_vcore_hour` is provided, it calculates the cost based on the total cores.
        Otherwise, it returns None.
        """
        if self.cost_per_vcore_hour is None:
            return None

        cost_per_hour = Decimal(self.get_total_cores()) * Decimal(self.cost_per_vcore_hour)
        job_cost = cost_per_hour * Decimal(duration_ms) / Decimal(3600000)  # Convert ms to hours
        return job_cost.quantize(Decimal('0.0000000000'))  # Ensure precision matches DECIMAL(18,10)
    
    def _convert_generic_to_specific_schema(self, generic_schema: list):
        """
        Convert a generic schema to a specific Spark schema.
        """
        import pyarrow as pa
        type_mapping = {
            'STRING': pa.string(),
            'TIMESTAMP': pa.timestamp('us', tz='UTC'),
            'TINYINT': pa.int8(),
            'SMALLINT': pa.int16(),
            'INT': pa.int32(),
            'BIGINT': pa.int64(),
            'FLOAT': pa.float32(),
            'DOUBLE': pa.float64(),
            'DECIMAL(18,10)': pa.decimal128(18, 10),
            'BOOLEAN': pa.bool_(),
            'MAP<STRING, STRING>': pa.map_(pa.string(), pa.string())
        }
        return pa.schema([(name, type_mapping[data_type]) for name, data_type in generic_schema])
    
    def _append_results_to_delta(self, abfss_path: str, results: list, generic_schema: list):
        """
        Appends a list of result records to an existing Delta table.

        Parameters
        ----------
        abfss_path : str
            The ABFSS URI or where the Delta table resides.
        results : list of dict
            A list of row dictionaries to append. Each dictionary may include an
            'engine_metadata' key, whose contents will be stored as a MAP<STRING,STRING>.
        generic_schema : list of tuple
            A list of (field_name, field_type) tuples defining the generic schema
            for the rows in `results`.

        Notes
        -----
        - Converts `generic_schema` into a PyArrow schema.
        - Extracts 'engine_metadata' from each row and appends it as a
          MAP<STRING,STRING> column.
        - Uses DeltaRs to write the data in "append" mode.
        - If the installed `deltalake` version is 0.x, forces the Rust engine.
        """
        import pyarrow as pa
        from ..engines.delta_rs import DeltaRs

        schema = self._convert_generic_to_specific_schema(generic_schema=generic_schema)

        # Extract engine_metadata and convert to map format, otherwise is interpreted as a Struct
        index = schema.get_field_index("engine_metadata")
        schema = schema.remove(index)
        map_data = []
        for result in results:
            metadata = result.pop('engine_metadata', {})
            if metadata:
                map_items = [(str(k), str(v)) for k, v in metadata.items()]
            else:
                map_items = []

            map_data.append(map_items)

        table = pa.Table.from_pylist(results, schema)
        map_array = pa.array(map_data, type=pa.map_(pa.string(), pa.string()))
        table = table.append_column('engine_metadata', map_array)

        if version('deltalake').startswith('0.'):
            DeltaRs().write_deltalake(
                abfss_path, 
                table, 
                mode="append",
                schema_mode='merge',
                engine='rust'
            )
        else:
            DeltaRs().write_deltalake(
                abfss_path, 
                table, 
                mode="append",
                schema_mode='merge'
            )


from .spark import Spark
from typing import Optional

try:
    from IPython.core.getipython import get_ipython
    utils = get_ipython().user_ns["mssparkutils"]
except Exception as e:
    e

class FabricSpark(Spark):
    """
    Spark Engine for ELT Benchmarks.
    """

    def __init__(
            self,
            lakehouse_name: str, 
            lakehouse_schema_name: str,
            spark_measure_telemetry: bool = False,
            cost_per_vcore_hour: Optional[float] = None
            ):
        """
        Initialize the SparkEngine with a Spark session.
        """
        self.lakehouse_name = lakehouse_name
        self.lakehouse_schema_name = lakehouse_schema_name

        super().__init__(catalog_name=self.lakehouse_name, schema_name=self.lakehouse_schema_name, spark_measure_telemetry=spark_measure_telemetry, cost_per_vcore_hour=cost_per_vcore_hour)

        self.version: str = f"{self.spark.sparkContext.version} (vhd_name=={self.spark.conf.get('spark.synapse.vhd.name')})"
        self.cost_per_vcore_hour = cost_per_vcore_hour or self._FABRIC_USD_COST_PER_VCORE_HOUR



from ....engines.spark import Spark

class SparkELTBench:
    def __init__(self, engine: Spark):
        
        import numpy as np
        self.np = np
        self.engine = engine
        self.engine.create_schema_if_not_exists(drop_before_create=True)

    def create_total_sales_fact(self):
        self.engine.spark.sql("""
            CREATE OR REPLACE TABLE total_sales_fact USING DELTA AS
            SELECT 
                s.s_store_id,
                i.i_item_id,
                c.c_customer_id,
                d.d_date AS sale_date,
                SUM(ss.ss_quantity) AS total_quantity,
                SUM(ss.ss_net_paid) AS total_net_paid,
                SUM(ss.ss_net_profit) AS total_net_profit
            FROM 
                store_sales ss
            JOIN 
                date_dim d ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN 
                store s ON ss.ss_store_sk = s.s_store_sk
            JOIN 
                item i ON ss.ss_item_sk = i.i_item_sk
            JOIN 
                customer c ON ss.ss_customer_sk = c.c_customer_sk
            WHERE 
                d.d_year = 2001
            GROUP BY 
                s.s_store_id, i.i_item_id, c.c_customer_id, d.d_date
            ORDER BY 
                s.s_store_id, d.d_date;
        """)

    def merge_percent_into_total_sales_fact(self, percent: float):
        seed = self.np.random.randint(1, high=1000, size=None, dtype=int)
        modulo = int(1 / percent)

        sampled_fact_data = self.engine.spark.sql(f"""
            SELECT 
                s.s_store_id, 
                i.i_item_id, 
                CASE 
                    WHEN rand() > 0.5 THEN 
                        CONCAT('NEW_', CAST(new_uid_val AS STRING))
                    ELSE 
                        c.c_customer_id
                END AS c_customer_id,
                d.d_date AS sale_date,
                ss.ss_quantity + FLOOR(rand() * 5 + 1) AS total_quantity,
                ss.ss_net_paid + rand() * 50 + 5 AS total_net_paid,
                ss.ss_net_profit + rand() * 20 + 1 AS total_net_profit
            FROM (
                SELECT *, 
                       ss_customer_sk + ss_sold_date_sk + {seed} AS new_uid_val
                FROM store_sales
                WHERE MOD(ss_customer_sk + ss_sold_date_sk + {seed}, {modulo}) = 0
            ) ss
            JOIN date_dim d 
              ON ss.ss_sold_date_sk = d.d_date_sk
            JOIN store s 
              ON ss.ss_store_sk = s.s_store_sk
            JOIN item i 
              ON ss.ss_item_sk = i.i_item_sk
            JOIN customer c 
              ON ss.ss_customer_sk = c.c_customer_sk
            """)

        # Register as a temporary view for SQL-based merge
        sampled_fact_data.createOrReplaceTempView("sampled_fact_data")

        self.engine.spark.sql("""
            MERGE INTO total_sales_fact AS target
            USING (
                SELECT 
                    s_store_id, 
                    i_item_id, 
                    c_customer_id, 
                    sale_date, 
                    total_quantity, 
                    total_net_paid, 
                    total_net_profit
                FROM sampled_fact_data
            ) AS source
            ON 
                target.s_store_id = source.s_store_id AND 
                target.i_item_id = source.i_item_id AND 
                target.c_customer_id = source.c_customer_id AND 
                target.sale_date = source.sale_date
            WHEN MATCHED THEN 
                UPDATE SET 
                    target.total_quantity = target.total_quantity + source.total_quantity,
                    target.total_net_paid = target.total_net_paid + source.total_net_paid,
                    target.total_net_profit = target.total_net_profit + source.total_net_profit
            WHEN NOT MATCHED THEN 
                INSERT (s_store_id, i_item_id, c_customer_id, sale_date, total_quantity, total_net_paid, total_net_profit)
                VALUES (source.s_store_id, source.i_item_id, source.c_customer_id, source.sale_date, source.total_quantity, source.total_net_paid, source.total_net_profit);
        """)
        
    def query_total_sales_fact(self):
        df = self.engine.spark.sql(f"""
                            select sum(total_net_profit), year(sale_date) 
                            from total_sales_fact group by year(sale_date)
                            """)
        result = df.collect()


from ....engines.spark import Spark

class SparkClickBench:
    def __init__(self, engine: Spark):
        
        self.engine = engine

    def load_parquet_to_delta(self, table_name: str, source_data_path: str):
        """
        Loads the ClickBench parquet data into Delta format using Spark.

        Parameters
        ----------
        source_data_path : str
            Path to the source parquet files.
        """
        from pyspark.sql import functions as sf
        # Load parquet files
        df = self.engine.spark.read.parquet(source_data_path)

        # ClickBench parquet data doesn't annotate the logical type of binary columns, therefore we cast as string.
        binary_cols = [c for c, t in df.dtypes if t == "binary"]
        df = df.withColumns({c: sf.col(c).cast("string") for c in binary_cols})
        # Datetime columns are stored as UNIX Integers and need to be converted explicity
        df = df.withColumn("EventTime", sf.col("EventTime").cast("timestamp"))
        df = df.withColumn("EventDate", sf.date_add(sf.lit("1970-01-01"), sf.col("EventDate")))
        df = df.withColumn("ClientEventTime", sf.col("ClientEventTime").cast("timestamp"))
        df = df.withColumn("LocalEventTime", sf.col("LocalEventTime").cast("timestamp"))

        df.write.format("delta").mode("append").saveAsTable(table_name)

    def execute_sql_query(self, query: str):
        return self.engine.execute_sql_query(query)


from abc import ABC, abstractmethod
from typing import Dict, Type, Optional
import uuid
from datetime import datetime
from ..utils.timer import timer
from ..engines.base import BaseEngine
from importlib.metadata import version, PackageNotFoundError

class BaseBenchmark(ABC):
    """
    Abstract base class for defining benchmarks. This class provides a structure for implementing benchmarks
    with a specific engine and scenario, and includes functionality for timing and saving results.

    Attributes
    ----------
    BENCHMARK_IMPL_REGISTRY : Dict[Type, Type]
        A registry for engines that the benchmark supports. If the engine requires a specific implementation
        that doesn't use the engines existing methods, the dictionary will map engines to the specific implementation
        class rather than. If only shared methods are used, the dictionary value will be None.
    engine : object
        The engine used to execute the benchmark.
    scenario_name : str
        The name of the scenario being benchmarked.
    result_abfss_path : Optional[str]
        The path where benchmark results will be saved, if `save_results` is True.
    save_results : bool
        Flag indicating whether to save benchmark results to a Delta table.
    header_detail_dict : dict
        A dictionary containing metadata about the benchmark run, including run ID, datetime, engine type,
        benchmark name, scenario name, total cores, and compute size.
    timer : object
        A timer object used to measure the duration of benchmark phases.
    results : list
        A list to store benchmark results.
        
    Methods
    -------
    run()
        Abstract method that must be implemented by subclasses to define the benchmark logic.
    post_results()
        Processes and saves benchmark results. If `save_results` is True, results are appended to a Delta table
        at the specified `result_abfss_path`. Clears the timer results after processing.
    """
    BENCHMARK_IMPL_REGISTRY: Dict[Type[BaseEngine], Type] = {}
    RESULT_SCHEMA = [
        ('run_id', 'STRING'),
        ('run_datetime', 'TIMESTAMP'),
        ('lakebench_version', 'STRING'),
        ('engine', 'STRING'),
        ('engine_version', 'STRING'),
        ('benchmark', 'STRING'),
        ('benchmark_version', 'STRING'),
        ('scale_factor', 'INT'),
        ('scenario', 'STRING'),
        ('total_cores', 'SMALLINT'),
        ('compute_size', 'STRING'),
        ('phase', 'STRING'),
        ('test_item', 'STRING'),
        ('start_datetime', 'TIMESTAMP'),
        ('duration_ms', 'INT'),
        ('estimated_job_cost', 'DECIMAL(18,10)'),
        ('iteration', 'TINYINT'),
        ('success', 'BOOLEAN'),
        ('error_message', 'STRING'),
        ('engine_metadata', 'MAP<STRING, STRING>')
    ]
    VERSION = ''

    def __init__(self, engine, scenario_name: str, result_abfss_path: Optional[str], save_results: bool = False):
        self.engine = engine
        self.scenario_name = scenario_name
        self.result_abfss_path = result_abfss_path
        self.save_results = save_results

        self.header_detail_dict = {
            'run_id': str(uuid.uuid1()),
            'run_datetime': datetime.now(),
            'lakebench_version': version('lakebench'),
            'engine': type(engine).__name__,
            'engine_version': self.engine.version,
            'benchmark': self.__class__.__name__,
            'benchmark_version': self.VERSION,
            'scale_factor': getattr(self, 'scale_factor', None),
            'scenario': scenario_name,
            'total_cores': self.engine.get_total_cores(),
            'compute_size': self.engine.get_compute_size()
        }
        self.timer = timer
        self.timer.clear_results()
        self.results = []

    @classmethod
    def register_engine(cls, engine_class: Type[BaseEngine], benchmark_impl: Optional[Type] = None):
        """
        Registers a custom engine class and its corresponding benchmark implementation.

        Parameters
        ----------
        engine_class : Type[BaseEngine]
            The engine class to register.
        benchmark_impl : Type[BaseBenchmark], optional
            The benchmark implementation class for the engine. If None, the engine's default methods will be used.
        """
        cls.BENCHMARK_IMPL_REGISTRY[engine_class] = benchmark_impl

    @abstractmethod
    def run(self):
        pass

    def post_results(self):
        """
        Processes and posts benchmark results, saving them to a specified location if save_results is True.
        This method collects timing results from the benchmark execution, formats them into a 
        structured array, and optionally saves the results to a Delta table. It also clears the timer 
        instance after offloading results to the `self.results` attribute.

        Parameters
        ----------
        None
        
        Notes
        -----
        - If `save_results` is True, the results are appended to the Delta table specified by 
          `result_abfss_path` using the `engine.append_array_to_delta` method.
        - After processing, the results are stored in `self.results` and the timer results are cleared.
        
        Examples
        --------
        >>> benchmark = Benchmark()
        >>> benchmark.post_results()
        # Processes the results and saves them if `save_results` is True.
        # post_results() should be called after each major benchmark phase.
        """

        result_array = [
            {
                **self.header_detail_dict,
                'phase': phase,
                'test_item': test_item,
                'start_datetime': start_datetime,
                'duration_ms': duration_ms,
                'estimated_job_cost': self.engine.get_job_cost(duration_ms), 
                'iteration': iteration,
                'success': success,
                'error_message': error_message
            }
            for phase, test_item, start_datetime, duration_ms, iteration, success, error_message in self.timer.results
        ]
        self.results.extend(result_array)

        if self.save_results:
            if self.result_abfss_path is None:
                raise ValueError("result_abfss_path must be provided if save_results is True.")
            else:
                try:
                    self.engine._append_results_to_delta(self.result_abfss_path, result_array, self.RESULT_SCHEMA)
                except Exception as e:
                    raise e
                finally:
                    self.timer.clear_results()



from ....engines.polars import Polars
from ....engines.delta_rs import DeltaRs

from IPython.core.getipython import get_ipython
notebookutils = get_ipython().user_ns.get("notebookutils")

import posixpath

class PolarsELTBench:
    def __init__(self, engine: Polars):

        import numpy as np
        self.np = np
        self.delta_rs = DeltaRs()
        self.write_deltalake = self.delta_rs.write_deltalake
        self.DeltaTable = self.delta_rs.DeltaTable
        self.storage_options={
            "bearer_token": notebookutils.credentials.getToken('storage')
        }
        self.engine = engine

    def create_total_sales_fact(self):
        fact_table_df = (
            self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'store_sales'), storage_options=self.storage_options)
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'date_dim'), storage_options=self.storage_options), left_on="ss_sold_date_sk", right_on="d_date_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'store'), storage_options=self.storage_options), left_on="ss_store_sk", right_on="s_store_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'item'), storage_options=self.storage_options), left_on="ss_item_sk", right_on="i_item_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'customer'), storage_options=self.storage_options), left_on="ss_customer_sk", right_on="c_customer_sk"
            )
            .with_columns(
                    self.engine.pl.col("d_date").alias("sale_date")
                )
            .filter(self.engine.pl.col("d_year") == 2001)
            .group_by(["s_store_id", "i_item_id", "c_customer_id", "sale_date"])
            .agg([
                self.engine.pl.sum("ss_quantity").alias("total_quantity"),
                self.engine.pl.sum("ss_net_paid").alias("total_net_paid"),
                self.engine.pl.sum("ss_net_profit").alias("total_net_profit")
            ])
            .sort(["s_store_id", "sale_date"])
        )

        fact_table_df.collect(engine='streaming').write_delta(
            posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'),
            mode="overwrite",
            storage_options=self.storage_options
        )


    def merge_percent_into_total_sales_fact(self, percent: float):
        seed = self.np.random.randint(1, high=1000, size=None, dtype=int)
        modulo = int(1 / percent)
        sampled_fact_data = (
            self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'store_sales'), storage_options=self.storage_options)
            .filter(
                ((self.engine.pl.col("ss_item_sk") * 1000000 + self.engine.pl.col("ss_ticket_number") + seed).hash() % modulo) == 0
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'date_dim'), storage_options=self.storage_options), 
                left_on="ss_sold_date_sk", right_on="d_date_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'store'), storage_options=self.storage_options), 
                left_on="ss_store_sk", right_on="s_store_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'item'), storage_options=self.storage_options), 
                left_on="ss_item_sk", right_on="i_item_sk"
            )
            .join(
                self.engine.pl.scan_delta(posixpath.join(self.engine.delta_abfss_schema_path, 'customer'), storage_options=self.storage_options), 
                left_on="ss_customer_sk", right_on="c_customer_sk"
            )
            .with_columns([
                # Create hash-based pseudo-random values for each row
                (self.engine.pl.col("ss_customer_sk") + self.engine.pl.col("ss_sold_date_sk") + seed).alias("new_uid_val")
            ])
            .filter(
                (self.engine.pl.col("new_uid_val") % modulo) == 0
            )
            .with_columns([
                self.engine.pl.col("s_store_id"),
                self.engine.pl.col("i_item_id"),
                self.engine.pl.when(self.engine.pl.col("new_uid_val") % 2 == 0)
                    .then(self.engine.pl.col("c_customer_id"))
                    .otherwise(self.engine.pl.concat_str([self.engine.pl.lit('NEW_'), self.engine.pl.col("new_uid_val")], separator=''))
                    .alias("c_customer_id"),
                self.engine.pl.col("d_date").alias("sale_date"),
                (self.engine.pl.col("ss_quantity") + (self.engine.pl.col("new_uid_val") % 5) + 1).alias("total_quantity"),
                (self.engine.pl.col("ss_net_paid") + ((self.engine.pl.col("new_uid_val") % 5000) / 100.0) + 5).alias("total_net_paid"),
                (self.engine.pl.col("ss_net_profit") + ((self.engine.pl.col("new_uid_val") % 2000) / 100.0) + 1).alias("total_net_profit")
            ])
            .select([
                "s_store_id",
                "i_item_id", 
                "c_customer_id",
                "sale_date",
                "total_quantity",
                "total_net_paid",
                "total_net_profit"
            ])
        )

        sampled_fact_data.collect(engine='streaming').write_delta(
            posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'), 
            mode="merge", 
            delta_merge_options={
                "predicate": """
                    target.s_store_id = source.s_store_id AND 
                    target.i_item_id = source.i_item_id AND 
                    target.c_customer_id = source.c_customer_id AND 
                    target.sale_date = source.sale_date
                    """,
                "source_alias": "source",
                "target_alias": "target"
            }, 
            storage_options=self.storage_options
        ) \
        .when_matched_update({
            "total_quantity": "target.total_quantity + source.total_quantity",
            "total_net_paid": "target.total_net_paid + source.total_net_paid",
            "total_net_profit": "target.total_net_profit + source.total_net_profit",
        }) \
        .when_not_matched_insert({
            "s_store_id": "source.s_store_id",
            "i_item_id": "source.i_item_id",
            "c_customer_id": "source.c_customer_id",
            "sale_date": "source.sale_date",
            "total_quantity": "source.total_quantity",
            "total_net_paid": "source.total_net_paid",
            "total_net_profit": "source.total_net_profit",
        }).execute()

    def query_total_sales_fact(self):
        query_df = self.engine.pl.scan_delta(
            posixpath.join(self.engine.delta_abfss_schema_path, 'total_sales_fact'), storage_options=self.storage_options
        ).group_by(
            self.engine.pl.col("sale_date").dt.year()
        ).agg(
            self.engine.pl.sum("total_net_profit").alias("sum_net_profit")
        ).collect()


from .base import BaseEngine

class DeltaRs(BaseEngine):
    """
    Delta-Rs Engine for ELT Benchmarks.
    """

    def __init__(self):
        """
        Initialize the Delta-rs Engine Configs
        """
        from deltalake.writer import write_deltalake
        from deltalake import DeltaTable
        self.write_deltalake = write_deltalake
        self.DeltaTable = DeltaTable
        